/*******************************************************************************
 * Questionnaire runtime
 *
 * Copyright (C) 2015 martin.guist@eoda.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package de.tinu.qtafi4;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FilenameFilter;
import java.io.InputStream;
import java.sql.Connection;
import java.sql.Date;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import javax.sql.DataSource;

/**
 * 
 * @author maks
 *
 */
public class DbProjectConfiguration implements ProjectConfiguration {

	public static final int STATUS_OFFLINE = 0;
	public static final int STATUS_ONLINE = 1;
	
	final String name;
	final DataSource dataSource;
	private AuthenticationType authentication = AuthenticationType.PIN;
	private Runtime runtime = Runtime.SETUP;
	private LogMode logmode = LogMode.OFF;
	private String rootpath;
	
	public DbProjectConfiguration(String name, DataSource ds, String rootpath) {
		this.name = name;
		this.dataSource = ds;
		this.rootpath = rootpath;
	}
	
	public void setRootPath(String root){
		this.rootpath = root;
	}
	
	@Override
	public void init() {
		Date now = new Date(System.currentTimeMillis());
		Connection con = null;
		PreparedStatement pst = null;
		try {
			con = dataSource.getConnection();
			pst = con.prepareStatement("SELECT * FROM `__projects` WHERE name=?");
			pst.setString(1, this.name);
			ResultSet rs =  pst.executeQuery();
			if(rs.next()){
				Date startDate = rs.getDate("start");
				Date endDate = rs.getDate("end");
				/*
				if(startDate.after(now)) runtime = Runtime.SETUP;
				else if(endDate.before(now)) runtime = Runtime.TERMINATED;
				else runtime = Runtime.ONLINE;
				*/
				int status = rs.getInt("status");
				if(status == STATUS_ONLINE) runtime = Runtime.ONLINE;
				else runtime = runtime.SETUP;
				
				
				String auth = rs.getString("authentication");
				switch(auth.toLowerCase()){
					case "no": authentication = AuthenticationType.NO; break;
					case "pin": authentication = AuthenticationType.PIN; break;
					case "open": authentication = AuthenticationType.OPEN; break;
					case "cookie": authentication = AuthenticationType.COOKIE; break;
					case "url": authentication = AuthenticationType.URL; break;
					case "urlbase64": authentication = AuthenticationType.URLBASE64; break;
				}
				
				String log = rs.getString("log");
				switch(log.toLowerCase()){
					case "off": logmode = LogMode.OFF; break;
					case "verbose": logmode = LogMode.VERBOSE; break;
				}
			}
			pst.close();
		}
		catch(SQLException ex){
			ex.printStackTrace();
		}
		finally {
			try {
				con.close();
			} 
			catch (SQLException e) {
				e.printStackTrace();
			}
		}
	}

	@Override
	public String getProjectName() {
		return this.name;
	}

	@Override
	public boolean projectResourceExists(String file) {
		return new File(rootpath + "/" + name + "/" + file).exists();
	}

	@Override
	public List<String> listQuestionnaires() {
		File[] fs =  new File(rootpath + "/" + name + "/").listFiles(new FilenameFilter() {
			@Override
			public boolean accept(File dir, String name) {
				if(name.toLowerCase().endsWith(".qml")) return true;
				return false;
			}
		});
		List<String> files = new ArrayList<String>();
		for(File f: fs) files.add(f.getName());
		return files;
	}
	
	@Override
	public InputStream getProjectResource(String file) {
		try {
			return new FileInputStream(new File(rootpath + "/" + name + "/" + file));
		} 
		catch (FileNotFoundException e) {
			System.err.println("resource " + file + " not found!");
		}
		return null;
	}

	@Override
	public InputStream getConfigurationResource(String file) {
		try {
			return new FileInputStream(new File(rootpath + "/" + name + "/" + file));
		} 
		catch (FileNotFoundException e) {
			System.err.println("file " + file + " not found!");
		}
		return null;
	}

	@Override
	public AuthenticationType getAuthentication() {
		return authentication;
	}

	@Override
	public Runtime getRuntime() {
		return runtime;
	}

	@Override
	public LogMode getLogMode() {
		return logmode;
	}

}
