/*******************************************************************************
 * Questionnaire runtime
 *
 * Copyright (C) 2015 martin.guist@eoda.de
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package de.tinu.qtafi4;

import java.io.IOException;
import java.io.StringWriter;
import java.util.Enumeration;
import java.util.Properties;
import java.util.Set;

import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;
import javax.sql.DataSource;

import org.apache.velocity.Template;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;

import de.tinu.qtafi4.auth.Authentication;
import de.tinu.qtafi4.auth.AuthenticationError;
import de.tinu.qtafi4.auth.AuthenticationManager;
import de.tinu.qtafi4.auth.AuthenticationToken;
import de.tinu.qtafi4.data.AnonymUser;
import de.tinu.qtafi4.data.SurveyUser;
import de.tinu.qtafi4.data.User;
import de.tinu.qtafi4.log.IncidentType;
import de.tinu.qtafi4.log.Phase;
import de.tinu.qtafi4.log.SQLLog;
import de.tinu.qtafi4.persistence.SQLCommunicator;
import de.tinu.qtafi4.postprocessor.Postprocessor;
import de.tinu.qtafi4.qml.Preprocessor;
/**
 * 
 * @author maks
 *
 */
public class QMLProcessor {
	
	static final int MAX_INCIDENT_COUNT = 20;
	
	private final Preprocessor preprocessor;
	private final Postprocessor postprocessor;
	private DataSource dataSource;
	
	private SQLCommunicator communicator;
	private SQLLog log;
	private VelocityContext context = new VelocityContext();
	private AuthenticationManager authenticationmanager;
	
	public QMLProcessor(){
		preprocessor = new Preprocessor();
		postprocessor = new Postprocessor();
	}
	
	public void setDataSource(DataSource dataSource) {
		this.dataSource = dataSource;
	}
	
	public void init(FilterConfig config){
		Properties p = new Properties();
		p.setProperty("resource.loader", "file");
		p.setProperty("file.resource.loader.class", "org.apache.velocity.runtime.resource.loader.FileResourceLoader");
		p.setProperty("file.resource.loader.path", config.getServletContext().getRealPath(config.getServletContext().getInitParameter("template-path")));
		
		p.setProperty("input.encoding", "UTF-8");
		p.setProperty("output.encoding", "UTF-8");
		p.put("runtime.log.logsystem.class", "org.apache.velocity.runtime.log.SimpleLog4JLogSystem");
		p.put("runtime.log.logsystem.log4j.category", "velocity");
		p.put("runtime.log.logsystem.log4j.logger", "velocity");
		Velocity.init(p);
		
		communicator = new SQLCommunicator(dataSource);
		log = new SQLLog(dataSource);
		authenticationmanager = new AuthenticationManager(communicator);
	}
	
	public void doHttp(QMLRequest request, HttpServletResponse response) throws IOException, ServletException {
		request.getHttpRequest().setCharacterEncoding("utf-8");
		if(request.getError() == SurveyError.PROJECT_ERROR){
			doError(request, response, new AnonymUser(), new ErrorMessage("Not found"));
			return;
		}
		
		String projectname = request.getProjectConfiguration().getProjectName();
		communicator.init(projectname);
		log.init(projectname);
		log.logRequest(request, projectname);
		if(log.getIncidentCount(request, IncidentType.AUTHENTICATIONERROR, projectname) > MAX_INCIDENT_COUNT){
			doError(request, response, new AnonymUser(), new ErrorMessage("IP blocked"));
			return;
		}
		if(request.getError() == SurveyError.FILE_NOT_FOUND){
			doError(request, response, new AnonymUser(), new ErrorMessage("File not found"));
			return;
		}
		if(request.getProjectConfiguration().getRuntime() == Runtime.SETUP){
			//doSetup(request, response, new AnonymUser());
			Authentication authentication = authenticationmanager.getAuthentication(request);
			if(request.getHttpRequest().getQueryString() != null){
				if(request.getHttpRequest().getQueryString().length() > 1){
					doLogin(request, response, authentication);
					return;
				}
			}
			if((! authentication.isAuthenticated())){
				doAnonym(request, response, new AnonymUser());
				return;
			}
			SurveyUser user = (SurveyUser) authentication.getPrincipal();
			user.write(request);
			doPreview(request, response, user);
			return;
		}
	
		
		if(request.getProjectConfiguration().getRuntime() == Runtime.TERMINATED){
			doFinished(request, response, new AnonymUser());
			return;
		}
		
		Authentication authentication = authenticationmanager.getAuthentication(request);
		//System.out.println("ID = "+ request.getSessionId());
	
		
		//if(request.getHttpRequest().getParameter("login") != null){
		if(request.getHttpRequest().getQueryString() != null){
			if(request.getHttpRequest().getQueryString().length() > 1){
				authentication = authenticationmanager.newAuthentication(request);
				authenticationmanager.logout();
				//System.out.println("login");
				doLogin(request, response, authentication);
				return;
			}
		}
		
		if((! authentication.isAuthenticated())){
			doAnonym(request, response, new AnonymUser());
			return;
		}
		SurveyUser user = (SurveyUser) authentication.getPrincipal();
		user.write(request);
		doAuthenticated(request, response, user);
		communicator.registerColumns(request.getColumnRegistry(), request.getProjectConfiguration().getProjectName());
	}
	
	
	private void doLogin(QMLRequest request, HttpServletResponse response, Authentication authentication) throws IOException, ServletException {
		String projectname = request.getProjectConfiguration().getProjectName();
		AuthenticationToken token = new AuthenticationToken(request);
		try {
			authentication = authenticationmanager.newAuthentication(request);
			//System.out.println("login ");
			authenticationmanager.login(authentication, token);
			SurveyUser user = (SurveyUser) authentication.getPrincipal();
			response.sendRedirect(request.recodeUrl());
			return;
		}
		catch (AuthenticationError err) {
			if(request.isPostRequest()){
				request.setError(SurveyError.AUTHENTICATIONERROR);
				log.logIncident(request, IncidentType.AUTHENTICATIONERROR, Phase.AUTHENTICATION, projectname);
				//System.out.println("CredentialError");
			}
			doAnonym(request, response, new AnonymUser());
			return;
		}
	}

	public void doAsynchron(QMLRequest request, HttpServletResponse response) 
			throws IOException, ServletException {
		String projectname = request.getProjectConfiguration().getProjectName();
		response.setContentType("text/html");
		response.setCharacterEncoding("utf-8");
		if(log.getIncidentCount(request, IncidentType.AUTHENTICATIONERROR, projectname) > MAX_INCIDENT_COUNT){
			response.getWriter().append("Max incident count ERROR");
			return;
		}
		if(request.getProjectConfiguration().getRuntime() == Runtime.SETUP){
			response.getWriter().append("SETUP");
			return;
		}
		if(request.getProjectConfiguration().getRuntime() == Runtime.TERMINATED){
			response.getWriter().append("TERMINATED");
			return;
		}
		
		Authentication authentication = authenticationmanager.getAuthentication(request);
		if(! authentication.isAuthenticated()){
			response.getWriter().append("Authentication ERROR");
			return;
		}
		SurveyUser data = (SurveyUser)authentication.getPrincipal();
		data.write(request);
		String asyncid = request.getHttpRequest().getParameter("id");
		if(asyncid == null) response.getWriter().append("no async id ERROR");
		
		preprocessor.buildId(request, data, asyncid);
		
		
		String formatter = "main.xsl";
		if(request.getQMLDocument().pagedir.equals("rtl")) formatter = "main-rtl.xsl";
		
		request.getQMLDocument().transform(request.getHttpRequest().getServletContext().getRealPath("/WEB-INF/formatter/" + formatter),
				"locale", request.getLanguage(),
				"mode", request.getProjectConfiguration().getAuthentication().toString(),
				"action-path", request.recodeUrl(),
				"max-page", data.getValue("__maxpage").toString(),
				"page", data.getValue("__page").toString());
		
		communicator.read(data, request.getProjectConfiguration().getProjectName());
		postprocessor.process(request, data);
		
		request.getQMLDocument().transform(response.getWriter(), 
				request.getHttpRequest().getServletContext().getRealPath("/WEB-INF/formatter/isolate_id.xsl"),
				"element-id", asyncid);
	}
	
	
	public void doError(QMLRequest request, HttpServletResponse response, User data, ErrorMessage error) 
			throws IOException, ServletException {
		//System.out.println("doError");
		response.setContentType("text/html");
		response.setCharacterEncoding("utf-8");
		StringWriter sw = new StringWriter();
		if(error != null){
			sw.append(error.toString());
		}
		else sw.append("Unknown error");
		context.put("body", sw.getBuffer());
		context.put("preview", false);
		context.put("pagedir", request.getQMLDocument().pagedir);
		Template template = null;
		try {
			template = Velocity.getTemplate("template.vm");
			template.merge(context, response.getWriter());
		}
		catch(Exception e){}
	}
	
	private void doFinished(QMLRequest request, HttpServletResponse response, User data) 
			throws IOException, ServletException {
		response.setContentType("text/html");
		response.setCharacterEncoding("utf-8");
		preprocessor.process(request, data);
		request.getQMLDocument().transform(request.getHttpRequest().getServletContext().getRealPath("/WEB-INF/formatter/finished.xsl"));
		postprocessor.process(request, data);
		StringWriter sw = new StringWriter();
		request.getQMLDocument().transform(sw, null);
		context.put("body", sw.getBuffer());
		context.put("pagedir", request.getQMLDocument().pagedir);
		Template template = null;
		try {
			template = Velocity.getTemplate("template.vm");
			template.merge(context, response.getWriter());
		}
		catch(Exception e){}
	}
	
	private void doSetup(QMLRequest request, HttpServletResponse response, User data) 
			throws IOException, ServletException {
		response.setContentType("text/html");
		response.setCharacterEncoding("utf-8");
		preprocessor.process(request, data);
		request.getQMLDocument().transform(request.getHttpRequest().getServletContext().getRealPath("/WEB-INF/formatter/setup.xsl"));
		postprocessor.process(request, data);
		StringWriter sw = new StringWriter();
		request.getQMLDocument().transform(sw, null);
		context.put("body", sw.getBuffer());
		context.put("preview", false);
		context.put("pagedir", request.getQMLDocument().pagedir);
		Template template = null;
		try {
			template = Velocity.getTemplate("template.vm");
			template.merge(context, response.getWriter());
		}
		catch(Exception e){}
	}
	
	/**
	 * Render first page
	 * 
	 */
	private void doAnonym(QMLRequest request, HttpServletResponse response, User data) 
			throws IOException, ServletException {
		//System.out.println("doAnonym");
		response.setCharacterEncoding("utf-8");
		response.setContentType("text/html");
		preprocessor.process(request, data);
		/*
		request.getQMLDocument().transform(System.out,null, "locale", request.getLanguage(),
				"action-path", request.recodeUrl(),
				"max-page", data.getValue("__maxpage").toString(),
				"page", data.getValue("__page").toString());
		*/
		
		String formatter = "main.xsl";
		if(request.getQMLDocument().pagedir.equals("rtl")) formatter = "main-rtl.xsl";
		
		request.getQMLDocument().transform(request.getHttpRequest().getServletContext().getRealPath("/WEB-INF/formatter/" + formatter),
				"locale", request.getLanguage(),
				"mode", request.getProjectConfiguration().getAuthentication().toString(),
				"action-path", request.recodeUrl(),
				"max-page", data.getValue("__maxpage").toString(),
				"page", data.getValue("__page").toString());
		
		postprocessor.process(request, data);
		StringWriter sw = new StringWriter();
		request.getQMLDocument().transform(sw, null);
		context.put("body", sw.getBuffer());
		if(request.getProjectConfiguration().getRuntime() == Runtime.SETUP){
			context.put("preview", true);
		}
		else context.put("preview", false);
		context.put("pagedir", request.getQMLDocument().pagedir);
		Template template = null;
		try {
			template = Velocity.getTemplate("template.vm", "UTF-8");
			template.merge(context, response.getWriter());
		}
		catch(Exception e){ e.printStackTrace();}
	}
	
	/**
	 * 
	 * @param request
	 * @param response
	 * @param data
	 * @throws IOException
	 * @throws ServletException
	 */
	private void doPreview(QMLRequest request, HttpServletResponse response, User data) 
			throws IOException, ServletException {
		//System.out.println("doPreview");
		HttpSession session = request.getHttpRequest().getSession();
		//System.out.println("session " + request.getSessionId());
		writeSessionData(session, data);
		response.setContentType("text/html");
		response.setCharacterEncoding("utf-8");
		readSessionData(session, data);
		preprocessor.process(request, data);
		//request.getQMLDocument().printDocument(System.out);
		String formatter = "main.xsl";
		if(request.getQMLDocument().pagedir.equals("rtl")) formatter = "main-rtl.xsl";
		
		request.getQMLDocument().transform(request.getHttpRequest().getServletContext().getRealPath("/WEB-INF/formatter/" + formatter),
				"locale", request.getLanguage(),
				"mode", request.getProjectConfiguration().getAuthentication().toString(),
				"action-path", request.recodeUrl(),
				"max-page", data.getValue("__maxpage").toString(),
				"page", Integer.toString(data.getIntValue("__page")));
		
		data.setValue("__timestamp", System.currentTimeMillis());
		postprocessor.process(request, data);
		StringWriter sw = new StringWriter();
		request.getQMLDocument().transform(sw, null);
		//System.out.println(sw.getBuffer());
		context.put("body", sw.getBuffer());
		context.put("preview", true);
		context.put("pagedir", request.getQMLDocument().pagedir);
		Template template = null;
		try {
			template = Velocity.getTemplate("template.vm", "UTF-8");
			template.merge(context, response.getWriter());
		}
		catch(Exception e){	}
	}
	
	private void writeSessionData(HttpSession session, User user){
		//System.out.println("writeSessionData");
		Set<String> keys = user.toMap().keySet();
		for(String ke: keys){
			String val = user.getValue(ke).toString();
			//System.out.println("-> " + ke + ": " + val);
			session.setAttribute(ke, val);
		}
	}
	
	private void readSessionData(HttpSession session, User user){
		//System.out.println("readSessionData");
		Enumeration<String> keys = session.getAttributeNames();
		for(; keys.hasMoreElements(); ){
			String key = keys.nextElement();
			String val = session.getAttribute(key).toString();
			//System.out.println("<- " + key + ": " + val);
			user.setValue(key, val);
		}
	}
	
	
	/**
	 * 
	 * @param request
	 * @param response
	 * @param data
	 * @throws IOException
	 * @throws ServletException
	 */
	private void doAuthenticated(QMLRequest request, HttpServletResponse response, User data) 
			throws IOException, ServletException {
		//System.out.println("doAuthenticated");
		communicator.write(data, request.getProjectConfiguration().getProjectName());
		response.setContentType("text/html");
		response.setCharacterEncoding("utf-8");
		
		//read database to fetch request page
		communicator.read(data, request.getProjectConfiguration().getProjectName());
		preprocessor.process(request, data);
		
		//request.getQMLDocument().printDocument(System.out);
		String formatter = "main.xsl";
		if(request.getQMLDocument().pagedir.equals("rtl")) formatter = "main-rtl.xsl";
		
		request.getQMLDocument().transform(request.getHttpRequest().getServletContext().getRealPath("/WEB-INF/formatter/" + formatter),
				"locale", request.getLanguage(),
				"mode", request.getProjectConfiguration().getAuthentication().toString(),
				"action-path", request.recodeUrl(),
				"max-page", data.getValue("__maxpage").toString(),
				"page", Integer.toString(data.getIntValue("__page")));
		
		data.setValue("__timestamp", System.currentTimeMillis());
		postprocessor.process(request, data);
		StringWriter sw = new StringWriter();
		request.getQMLDocument().transform(sw, null);
		//System.out.println(sw.getBuffer());
		context.put("body", sw.getBuffer());
		context.put("preview", false);
		context.put("pagedir", request.getQMLDocument().pagedir);
		Template template = null;
		try {
			template = Velocity.getTemplate("template.vm", "UTF-8");
			template.merge(context, response.getWriter());
		}
		catch(Exception e){	}
		
	}
}
